from brainsets.core import StringIntEnum


class Orientation_8_Classes(StringIntEnum):
    angle_0 = 0
    angle_45 = 1
    angle_90 = 2
    angle_135 = 3
    angle_180 = 4
    angle_225 = 5
    angle_270 = 6
    angle_315 = 7
